@@ -1,7 +1,6 @@
 import os
 from queue import Queue
 import requests
-import sys
 import threading
 import time
 from typing import List, Type
@@ -13,7 +12,7 @@ from uldlib.page import Page
 from uldlib.part import DownloadPart
 from uldlib.segfile import SegFileLoader
 from uldlib.torrunner import TorRunner
-from uldlib.utils import LogLevel
+from uldlib.utils import DownloaderError, DownloaderStopped, LogLevel
 
 
 class Downloader:
@@ -32,36 +31,60 @@ class Downloader:
 
     download_url_queue: Queue
     parts: int
+    tor: TorRunner
+    page: Page
 
-    def __init__(self, frontend: Type[Frontend], captcha_solver: Type[CaptchaSolver]):
+    def __init__(self, tor: TorRunner, frontend: Type[Frontend], captcha_solver: Type[CaptchaSolver]):
+        """Initialize the Downloader.
+
+           The TorRunner could be launched or not, the .launch() method will be called when needed.
+           Also it is caller responsibility to call .stop() method on the TorRunner.
+        """
+
+        self.success = None
+        self.url = None
+        self.filename = None
+        self.output_filename = None
+        self.stat_filename = None
+        self.total_size = None
         self.frontend = frontend
         self.log = frontend.main_log
         self.captcha_solver = captcha_solver
 
         self.cli_initialized = False
         self.conn_timeout = None
+        self.tor = tor
 
-        self.stop_download = threading.Event()
-        self.stop_captcha = threading.Event()
-        self.stop_frontend = threading.Event()
-
-    def terminate(self):
+    def terminate(self, quiet: bool = False):
         if self.terminating:
             return
         self.terminating = True
 
-        self.log('Terminating download. Please wait for stopping all threads.')
-        self.stop_captcha.set()
+        if not quiet:
+            self.log('Terminating download. Please wait for stopping all threads.', level=LogLevel.WARNING)
+
         self.stop_download.set()
-        if self.captcha_thread:
+        self.stop_captcha.set()
+        if self.captcha_thread and self.captcha_thread.is_alive():
             self.captcha_thread.join()
         for p in self.threads:
-            p.join()
-        self.log('Download terminated.')
+            if p.is_alive():
+                p.join()
+
+        if not quiet:
+            self.log('Download terminated.', level=LogLevel.WARNING)
+
         self.stop_frontend.set()
-        if self.frontend_thread:
+        if self.frontend_thread and self.frontend_thread.is_alive():
             self.frontend_thread.join()
 
+    def clean(self):
+        # remove resume .udown file
+        if os.path.exists(self.stat_filename):
+            os.remove(self.stat_filename)
+        if self.page.linkCache is not None:
+            self.page.linkCache.clean()
+
     def _captcha_breaker(self, page, parts):
         msg = ""
         if page.isDirectDownload:
@@ -69,11 +92,14 @@ class Downloader:
         else:
             msg = "Solve CAPTCHA dlink .."
 
-        for url in self.captcha_download_links_generator:
-            if self.stop_captcha.is_set():
-                break
-            self.captcha_solver.log(msg)
-            self.download_url_queue.put(url)
+        try:
+            for url in self.captcha_download_links_generator:
+                if self.stop_captcha.is_set():
+                    break
+                self.captcha_solver.log(msg)
+                self.download_url_queue.put(url)
+        except DownloaderError as e:
+            self.captcha_solver.log(str(e), level=LogLevel.ERROR)
 
     def _download_part(self, part: DownloadPart):
         try:
@@ -107,12 +133,15 @@ class Downloader:
                 "Connection": "close",
             })
 
-            if r.status_code != 429:
+            if r.status_code == 429:
+                part.set_status("Status code 429 Too Many Requests returned… will try again in few seconds", warning=True)
+                time.sleep(5)
+            elif r.status_code == 425:
+                part.set_status("Status code 425 Too Early returned… will try again in few seconds", warning=True)
+                time.sleep(5)
+            else:
                 break
 
-            part.set_status("Status code 429 Too Many Requests returned… will try again in few seconds", warning=True)
-            time.sleep(5)
-
         if r.status_code != 206 and r.status_code != 200:
             part.set_status(f"Status code {r.status_code} returned: {writer.pfrom + writer.written}/{writer.pto}", error=True)
             return
@@ -147,16 +176,17 @@ class Downloader:
         # reuse download link if need
         self.download_url_queue.put(part.download_url)
 
-    def download(self, url, parts=10, target_dir="", conn_timeout=DEFAULT_CONN_TIMEOUT):
+    def download(self, url: str, parts: int = 10, target_dir: str = "", temp_dir: str = "", do_overwrite: bool = False, conn_timeout=DEFAULT_CONN_TIMEOUT):
         """Download file from Uloz.to using multiple parallel downloads.
             Arguments:
-                url (str): URL of the Uloz.to file to download
-                parts (int): Number of parts that will be downloaded in parallel (default: 10)
-                target_dir (str): Directory where the download should be saved (default: current directory)
+                url: URL of the Uloz.to file to download
+                parts: Number of parts that will be downloaded in parallel (default: 10)
+                target_dir: Directory where the download should be saved (default: current directory)
+                do_overwrite: Overwrite files without asking
+                temp_dir: Directory where temporary files will be created (default: current directory)
         """
         self.url = url
         self.parts = parts
-        self.target_dir = target_dir
         self.conn_timeout = conn_timeout
 
         self.threads = []
@@ -164,29 +194,37 @@ class Downloader:
         self.isLimited = False
         self.isCaptcha = False
 
+        self.stop_download = threading.Event()
+        self.stop_captcha = threading.Event()
+        self.stop_frontend = threading.Event()
+
         # 1. Prepare downloads
         self.log("Starting downloading for url '{}'".format(url))
         # 1.1 Get all needed information
         self.log("Getting info (filename, filesize, …)")
 
         try:
-            tor = TorRunner()
-            page = Page(url, target_dir, parts, tor, self.conn_timeout)
+            self.page = Page(url, temp_dir, parts, self.tor, self.conn_timeout)
+            page = self.page  # shortcut
             page.parse()
 
-        except RuntimeError as e:
-            self.log('Cannot download file: ' + str(e), error=True)
-            sys.exit(1)
+        except Exception as e:
+            raise DownloaderError('Cannot download file: ' + str(e))
 
         # Do check - only if .udown status file not exists get question
-        output_filename = os.path.join(target_dir, page.filename)
-        if os.path.isfile(output_filename) and not os.path.isfile(output_filename+DOWNPOSTFIX):
-            answer = self.frontend.prompt(
-                "WARNING: File '{}' already exists, overwrite it? [y/n] ".format(output_filename),
-                level=LogLevel.WARNING
-            )
-            if answer != 'y':
-                sys.exit(1)
+        self.output_filename = os.path.join(target_dir, page.filename)
+        self.stat_filename = os.path.join(temp_dir, page.filename + DOWNPOSTFIX)
+        self.filename = page.filename
+        # .udown file is always present in cli_mode = False
+        if os.path.isfile(self.output_filename) and not os.path.isfile(self.stat_filename):
+            if self.frontend.supports_prompt and not do_overwrite:
+                answer = self.frontend.prompt(
+                    "WARNING: File '{}' already exists, overwrite it? [y/n] ".format(self.output_filename), level=LogLevel.WARNING)
+                if answer != 'y':
+                    raise DownloaderStopped()
+            else:
+                self.log("WARNING: File '{}' already exists, but .udown file not present. File will be overwritten.."
+                         .format(self.output_filename), level=LogLevel.WARNING)
 
         info = DownloadInfo()
         info.filename = page.filename
@@ -209,8 +247,7 @@ class Downloader:
                 info.download_type = "CAPTCHA protected download"
 
             if self.isCaptcha and self.captcha_solver.cannot_solve:
-                self.log("Cannot solve CAPTCHAs, no solver available. Terminating", level=LogLevel.ERROR)
-                sys.exit(1)
+                raise DownloaderError("Cannot solve CAPTCHAs, no solver available. Terminating")
 
             self.captcha_download_links_generator = page.captcha_download_links_generator(
                 solver=self.captcha_solver, stop_event=self.stop_captcha,
@@ -218,16 +255,15 @@ class Downloader:
             download_url = next(self.captcha_download_links_generator)
 
         head = requests.head(download_url, allow_redirects=True)
-        total_size = int(head.headers['Content-Length'])
+        self.total_size = int(head.headers['Content-Length'])
 
         try:
-            file_data = SegFileLoader(output_filename, total_size, parts)
+            file_data = SegFileLoader(self.output_filename, self.stat_filename, self.total_size, parts)
             writers = file_data.make_writers()
         except Exception as e:
-            self.log(f"Failed: Can not create '{output_filename}' error: {e} ", level=LogLevel.ERROR)
-            sys.exit(1)
+            raise DownloaderError(f"Failed: Can not create '{self.output_filename}' error: {e} ")
 
-        info.total_size = total_size
+        info.total_size = self.total_size
         info.part_size = file_data.part_size
         info.parts = file_data.parts
 
@@ -305,16 +341,11 @@ class Downloader:
             if part.error:
                 success = False
 
-        self.stop_captcha.set()
-        self.stop_frontend.set()
-        if self.captcha_thread:
-            self.captcha_thread.join()
-        if self.frontend_thread:
-            self.frontend_thread.join()
+        self.terminate(quiet=True)
 
+        self.success = success
         # result end status
         if not success:
-            self.log("Failure of one or more downloads, exiting", level=LogLevel.ERROR)
-            sys.exit(1)
+            raise DownloaderError("Failure of one or more downloads, exiting")
 
         self.log("All downloads successfully finished", level=LogLevel.SUCCESS)
