@@ -1,13 +1,15 @@
 from abc import abstractmethod
-from datetime import timedelta
+from datetime import datetime, timedelta
+from io import TextIOWrapper
 from traceback import print_exc
 import colors
 import os
 import sys
 import time
 import threading
-from typing import Dict, List, Tuple
+from typing import Dict, List, Optional, Tuple
 
+from uldlib import utils
 from uldlib.const import CLI_STATUS_STARTLINE
 from uldlib.part import DownloadPart
 from uldlib.utils import LogLevel
@@ -23,60 +25,102 @@ class DownloadInfo:
 
 
 class Frontend():
+    supports_prompt: bool
+
+    def __init__(self, supports_prompt: bool):
+        self.supports_prompt = supports_prompt
+
+    # Log functions for TOR, CAPTCHA solver and for everything else
+    # - `progress` is used by some frontends, when True it means that this message
+    #   is progress (progress bar or similar) and should not be logged / should be overwritten
 
     @abstractmethod
-    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO):
+    def tor_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
+        pass
+
+    @abstractmethod
+    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
+        pass
+
+    @abstractmethod
+    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
         pass
 
     @abstractmethod
     def captcha_stats(self, stats: Dict[str, int]):
         pass
 
-    @abstractmethod
-    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO):
-        pass
-
     @abstractmethod
     def prompt(self, msg: str, level: LogLevel = LogLevel.INFO) -> str:
         pass
 
     @abstractmethod
-    def run(self, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
+    def run(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
         pass
 
 
 class ConsoleFrontend(Frontend):
     cli_initialized: bool
+    show_parts: bool
+    logfile: Optional[TextIOWrapper] = None
 
     last_log: Tuple[str, LogLevel]
 
     last_captcha_log: Tuple[str, LogLevel]
     last_captcha_stats: Dict[str, int]
 
-    def __init__(self):
+    def __init__(self, show_parts: bool = False, logfile: str = ""):
+        super().__init__(supports_prompt=True)
         self.cli_initialized = False
         self.last_log = ("", LogLevel.INFO)
         self.last_captcha_log = ("", LogLevel.INFO)
         self.last_captcha_stats = None
+        self.show_parts = show_parts
+        if logfile:
+            self.logfile = open(logfile, 'a')
 
-    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO):
-        self.last_captcha_log = (msg, level)
+    def __del__(self):
+        if self.logfile:
+            self.logfile.close()
+
+    @staticmethod
+    def _log_print(msg: str, progress: bool):
+        if progress:
+            sys.stdout.write(msg + "\033[K\r")
+        else:
+            print(msg)
+
+    def _log_logfile(self, prefix: str, msg: str, progress: bool, level: LogLevel):
+        if progress or self.logfile is None:
+            return
+
+        t = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+        self.logfile.write(f"{t} {prefix}\t[{level.name}] {msg}\n")
+        self.logfile.flush()
+
+    def tor_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
+        self.last_captcha_log = (msg, level)  # shares same log with CAPTCHA
+        self._log_logfile('TOR ', msg, progress=progress, level=level)
         if not self.cli_initialized:
-            sys.stdout.write(colors.blue(
-                "[Link solve]\t") + self._color(msg, level) + "\033[K\r")
+            self._log_print(colors.blue("[TOR]\t") + utils.color(msg, level), progress=progress)
+
+    def captcha_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
+        self.last_captcha_log = (msg, level)
+        self._log_logfile('CAPTCHA', msg, progress=progress, level=level)
+        if not self.cli_initialized:
+            self._log_print(colors.blue("[Link solve]\t") + utils.color(msg, level), progress=progress)
+
+    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO, progress: bool = False):
+        self.last_log = (msg, level)
+        self._log_logfile('MAIN', msg, progress=progress, level=level)
+        if not self.cli_initialized:
+            self._log_print(utils.color(msg, level), progress=progress)
 
     def captcha_stats(self, stats: Dict[str, int]):
         self.last_captcha_stats = stats
 
-    def main_log(self, msg: str, level: LogLevel = LogLevel.INFO):
-        self.last_log = (msg, level)
-
-        if self.cli_initialized:
-            return
-        print(self._color(msg, level))
-
     def prompt(self, msg: str, level: LogLevel = LogLevel.INFO) -> str:
-        print(self._color(msg, level), end="")
+        print(utils.color(msg, level), end="")
         return input().strip()
 
     @staticmethod
@@ -96,16 +140,6 @@ class ConsoleFrontend(Frontend):
         sys.stdout.write(text)
         sys.stdout.flush()
 
-    @staticmethod
-    def _color(text: str, level: LogLevel) -> str:
-        if level == LogLevel.WARNING:
-            return colors.yellow(text)
-        if level == LogLevel.ERROR:
-            return colors.red(text)
-        if level == LogLevel.SUCCESS:
-            return colors.green(text)
-        return text
-
     def run(self, info: DownloadInfo, parts: List[DownloadPart], stop_event: threading.Event, terminate_func):
         try:
             self._loop(info, parts, stop_event)
@@ -141,31 +175,22 @@ class ConsoleFrontend(Frontend):
         y = 0
 
         while True:
-            if stop_event.is_set():
-                break
-
             t = time.time()
             # Get parts info
             lines = []
             s = 0
             for part in parts:
                 (line, level, size) = part.get_frontend_status()
-                lines.append(self._color(line, level))
+                lines.append(utils.color(line, level))
                 s += size
 
-            # Print parts
-            for (line, part) in zip(lines, parts):
-                self._print(
-                    colors.blue(f"[Part {part.id}]") + f"\t{line}",
-                    y=(part.id + CLI_STATUS_STARTLINE))
-
-            y = info.parts + CLI_STATUS_STARTLINE
+            y = CLI_STATUS_STARTLINE
 
             # Print CAPTCHA/TOR status
             (msg, level) = self.last_captcha_log
             self._print(
                 colors.yellow("[Link solve]\t") +
-                self._color(msg, level),
+                utils.color(msg, level),
                 y=y
             )
             y += 1
@@ -206,14 +231,25 @@ class ConsoleFrontend(Frontend):
             (msg, level) = self.last_log
             self._print(
                 colors.yellow("[STATUS]\t") +
-                self._color(msg, level),
+                (msg if level == LogLevel.INFO else colors.negative(utils.color(msg, level))),
                 y=y
             )
             y += 1
 
+            # Print parts
+            if self.show_parts:
+                for (line, part) in zip(lines, parts):
+                    self._print(
+                        colors.blue(f"[Part {part.id}]") + f"\t{line}",
+                        y=(y + part.id))
+
+            if stop_event.is_set():
+                break
+
             time.sleep(0.5)
 
         if self.cli_initialized:
+            y = info.parts + CLI_STATUS_STARTLINE + 4
             sys.stdout.write("\033[{};{}H".format(y + 2, 0))
             sys.stdout.write("\033[?25h")  # show cursor
             self.cli_initialized = False
